@@ -1,12 +1,14 @@
 from shutil import rmtree
 from os.path import join
 from os import makedirs
-from typing import Tuple
+from typing import Tuple, Type, Union, Dict, Any, TYPE_CHECKING
 
 import numpy as np
 
 from wkw import wkw
 
+if TYPE_CHECKING:
+    from wkcuber.api.Dataset import AbstractDataset, TiffDataset
 from wkcuber.api.MagDataset import (
     MagDataset,
     WKMagDataset,
@@ -23,23 +25,32 @@ class Layer:
     COLOR_TYPE = "color"
     SEGMENTATION_TYPE = "segmentation"
 
-    def __init__(self, name, dataset, dtype_per_channel, num_channels):
+    def __init__(
+        self,
+        name: str,
+        dataset: "AbstractDataset",
+        dtype_per_channel: np.dtype,
+        num_channels: int,
+    ) -> None:
         self.name = name
         self.dataset = dataset
         self.dtype_per_channel = dtype_per_channel
         self.num_channels = num_channels
-        self.mags = {}
+        self.mags: Dict[str, Any] = {}
 
         full_path = join(dataset.path, name)
         makedirs(full_path, exist_ok=True)
 
-    def get_mag(self, mag) -> MagDataset:
+    def get_mag(self, mag: Union[str, Mag]) -> MagDataset:
         mag = Mag(mag).to_layer_name()
         if mag not in self.mags.keys():
             raise IndexError("The mag {} is not a mag of this layer".format(mag))
         return self.mags[mag]
 
-    def delete_mag(self, mag):
+    def get_or_add_mag(self, mag: Union[str, Mag], **kwargs: Any) -> MagDataset:
+        pass
+
+    def delete_mag(self, mag: Union[str, Mag]) -> None:
         mag = Mag(mag).to_layer_name()
         if mag not in self.mags.keys():
             raise IndexError(
@@ -52,12 +63,12 @@ class Layer:
         full_path = find_mag_path_on_disk(self.dataset.path, self.name, mag)
         rmtree(full_path)
 
-    def _create_dir_for_mag(self, mag):
+    def _create_dir_for_mag(self, mag: Union[str, Mag]) -> None:
         mag = Mag(mag).to_layer_name()
         full_path = join(self.dataset.path, self.name, mag)
         makedirs(full_path, exist_ok=True)
 
-    def _assert_mag_does_not_exist_yet(self, mag):
+    def _assert_mag_does_not_exist_yet(self, mag: Union[str, Mag]) -> None:
         mag = Mag(mag).to_layer_name()
         if mag in self.mags.keys():
             raise IndexError(
@@ -68,31 +79,40 @@ class Layer:
 
     def set_bounding_box(
         self, offset: Tuple[int, int, int], size: Tuple[int, int, int]
-    ):
+    ) -> None:
         self.set_bounding_box_offset(offset)
         self.set_bounding_box_size(size)
 
-    def set_bounding_box_offset(self, offset: Tuple[int, int, int]):
-        size = self.dataset.properties.data_layers["color"].get_bounding_box_size()
-        self.dataset.properties._set_bounding_box_of_layer(
-            self.name, tuple(offset), tuple(size)
-        )
+    def set_bounding_box_offset(self, offset: Tuple[int, int, int]) -> None:
+        size: Tuple[int, int, int] = self.dataset.properties.data_layers[
+            "color"
+        ].get_bounding_box_size()
+        self.dataset.properties._set_bounding_box_of_layer(self.name, offset, size)
         for _, mag in self.mags.items():
             mag.view.global_offset = offset
 
-    def set_bounding_box_size(self, size: Tuple[int, int, int]):
-        offset = self.dataset.properties.data_layers["color"].get_bounding_box_offset()
-        self.dataset.properties._set_bounding_box_of_layer(
-            self.name, tuple(offset), tuple(size)
-        )
+    def set_bounding_box_size(self, size: Tuple[int, int, int]) -> None:
+        offset: Tuple[int, int, int] = self.dataset.properties.data_layers[
+            "color"
+        ].get_bounding_box_offset()
+        self.dataset.properties._set_bounding_box_of_layer(self.name, offset, size)
         for _, mag in self.mags.items():
             mag.view.size = size
 
+    def setup_mag(self, mag: str) -> None:
+        pass
+
 
 class WKLayer(Layer):
+    mags: Dict[str, WKMagDataset]
+
     def add_mag(
-        self, mag, block_len=None, file_len=None, block_type=None
-    ) -> WKMagDataset:
+        self,
+        mag: Union[str, Mag],
+        block_len: int = None,
+        file_len: int = None,
+        block_type: int = None,
+    ) -> MagDataset:
         if block_len is None:
             block_len = 32
         if file_len is None:
@@ -107,15 +127,18 @@ class WKLayer(Layer):
         self._create_dir_for_mag(mag)
 
         self.mags[mag] = WKMagDataset.create(self, mag, block_len, file_len, block_type)
-        self.dataset.properties._add_mag(self.name, mag, block_len * file_len)
+        self.dataset.properties._add_mag(
+            self.name, mag, cube_length=block_len * file_len
+        )
 
         return self.mags[mag]
 
-    def get_or_add_mag(
-        self, mag, block_len=None, file_len=None, block_type=None
-    ) -> WKMagDataset:
+    def get_or_add_mag(self, mag: Union[str, Mag], **kwargs: Any) -> MagDataset:
         # normalize the name of the mag
         mag = Mag(mag).to_layer_name()
+        block_len: int = kwargs.get("block_len", None)
+        file_len: int = kwargs.get("file_len", None)
+        block_type: int = kwargs.get("block_type", None)
 
         if mag in self.mags.keys():
             assert (
@@ -131,7 +154,7 @@ class WKLayer(Layer):
         else:
             return self.add_mag(mag, block_len, file_len, block_type)
 
-    def setup_mag(self, mag):
+    def setup_mag(self, mag: str) -> None:
         # This method is used to initialize the mag when opening the Dataset. This does not create e.g. the wk_header.
 
         # normalize the name of the mag
@@ -148,12 +171,14 @@ class WKLayer(Layer):
             self, mag, wk_header.block_len, wk_header.file_len, wk_header.block_type
         )
         self.dataset.properties._add_mag(
-            self.name, mag, wk_header.block_len * wk_header.file_len
+            self.name, mag, cube_length=wk_header.block_len * wk_header.file_len
         )
 
 
 class TiffLayer(Layer):
-    def add_mag(self, mag) -> MagDataset:
+    dataset: "TiffDataset"
+
+    def add_mag(self, mag: Union[str, Mag]) -> MagDataset:
         # normalize the name of the mag
         mag = Mag(mag).to_layer_name()
 
@@ -167,7 +192,7 @@ class TiffLayer(Layer):
 
         return self.mags[mag]
 
-    def get_or_add_mag(self, mag) -> MagDataset:
+    def get_or_add_mag(self, mag: Union[str, Mag], **kwargs: Any) -> MagDataset:
         # normalize the name of the mag
         mag = Mag(mag).to_layer_name()
 
@@ -176,7 +201,7 @@ class TiffLayer(Layer):
         else:
             return self.add_mag(mag)
 
-    def setup_mag(self, mag):
+    def setup_mag(self, mag: str) -> None:
         # This method is used to initialize the mag when opening the Dataset. This does not create e.g. folders.
 
         # normalize the name of the mag
@@ -189,10 +214,10 @@ class TiffLayer(Layer):
         )
         self.dataset.properties._add_mag(self.name, mag)
 
-    def _get_mag_dataset_class(self):
+    def _get_mag_dataset_class(self) -> Type[TiffMagDataset]:
         return TiffMagDataset
 
 
 class TiledTiffLayer(TiffLayer):
-    def _get_mag_dataset_class(self):
+    def _get_mag_dataset_class(self) -> Type[TiledTiffMagDataset]:
         return TiledTiffMagDataset
