@@ -1,15 +1,17 @@
+from typing import Any, Dict
+
 from wkcuber.mag import Mag
 
 
 class Resolution:
-    def __init__(self, mag):
+    def __init__(self, mag: str):
         self._mag = Mag(mag)
 
     def _to_json(self) -> dict:
         return {"resolution": self.mag.to_array()}
 
     @classmethod
-    def _from_json(cls, json_data):
+    def _from_json(cls, json_data: Dict[str, Any]) -> "Resolution":
         return cls(json_data["resolution"])
 
     @property
@@ -18,7 +20,7 @@ class Resolution:
 
 
 class WkResolution(Resolution):
-    def __init__(self, mag, cube_length):
+    def __init__(self, mag: str, cube_length: int):
         super().__init__(mag)
         self._cube_length = cube_length
 
@@ -26,7 +28,7 @@ class WkResolution(Resolution):
         return {"resolution": self.mag.to_array(), "cubeLength": self.cube_length}
 
     @classmethod
-    def _from_json(cls, json_data):
+    def _from_json(cls, json_data: Dict[str, Any]) -> "WkResolution":
         return cls(json_data["resolution"], json_data["cubeLength"])
 
     @property
