@@ -1,6 +1,7 @@
 import itertools
 import re
-from typing import Optional, List, Tuple, Set
+from types import TracebackType
+from typing import Optional, List, Tuple, Set, Type, Iterator, Sequence, cast, Union
 
 from skimage import io
 import numpy as np
@@ -49,7 +50,7 @@ def detect_tile_ranges(
         logger.info(
             f"Auto-detected tile ranges from tif directory structure: z {detected_z_range} x {detected_x_range} y {detected_y_range}"
         )
-        return (detected_z_range, detected_x_range, detected_y_range)
+        return detected_z_range, detected_x_range, detected_y_range
 
     raise Exception("Couldn't auto-detect tile ranges from wkw or tile path pattern")
 
@@ -60,7 +61,7 @@ def detect_tile_ranges_from_pattern_recursively(
     z_values: Set[int],
     x_values: Set[int],
     y_values: Set[int],
-) -> Tuple[Optional[range], Optional[range], Optional[range]]:
+) -> Tuple[range, range, range]:
     (
         current_pattern_element,
         prefix,
@@ -122,7 +123,7 @@ def advance_to_next_relevant_pattern_element(
     return current_pattern_element, prefix, remaining_pattern_elements
 
 
-def values_to_range(values: Set[int]):
+def values_to_range(values: Set[int]) -> range:
     if len(values) > 0:
         return range(min(values), max(values) + 1)
     return range(0, 0)
@@ -152,7 +153,9 @@ def detect_value(
     return []
 
 
-def to_file_name(pattern, x, y, z) -> str:
+def to_file_name(
+    pattern: str, x: Optional[int], y: Optional[int], z: Optional[int]
+) -> str:
     file_name = pattern
     if x is not None:
         file_name = replace_coordinate(file_name, "x", x)
@@ -163,17 +166,39 @@ def to_file_name(pattern, x, y, z) -> str:
     return file_name
 
 
+class TiffMagHeader:
+    def __init__(
+        self,
+        pattern: str = "{zzzzz}.tif",
+        dtype_per_channel: np.dtype = np.dtype("uint8"),
+        num_channels: int = 1,
+        tile_size: Optional[Tuple[int, int]] = (32, 32),
+    ) -> None:
+        self.pattern = pattern
+        self.dtype_per_channel = np.dtype(dtype_per_channel)
+        self.num_channels = num_channels
+        self.tile_size = tile_size
+
+
 class TiffMag:
-    def __init__(self, root, header):
+    def __init__(self, root: str, header: TiffMagHeader) -> None:
 
         self.root = root
         self.tiffs = dict()
         self.header = header
 
-        z_range, x_range, y_range = detect_tile_ranges(self.root, self.header.pattern)
-        z_range = [None] if z_range == range(0, 0) else z_range
-        y_range = [None] if y_range == range(0, 0) else y_range
-        x_range = [None] if x_range == range(0, 0) else x_range
+        detected_z_range, detected_x_range, detected_y_range = detect_tile_ranges(
+            self.root, self.header.pattern
+        )
+        z_range: List[Optional[int]] = (
+            [None] if detected_z_range == range(0, 0) else list(detected_z_range)
+        )
+        y_range: List[Optional[int]] = (
+            [None] if detected_y_range == range(0, 0) else list(detected_y_range)
+        )
+        x_range: List[Optional[int]] = (
+            [None] if detected_x_range == range(0, 0) else list(detected_x_range)
+        )
         available_tiffs = list(itertools.product(x_range, y_range, z_range))
 
         for xyz in available_tiffs:
@@ -182,18 +207,19 @@ class TiffMag:
                     self.get_file_name_for_layer(xyz)
                 )  # open is lazy
 
-    def read(self, off, shape) -> np.array:
+    def read(self, off: Tuple[int, int, int], shape: Tuple[int, int, int]) -> np.array:
         # modify the shape to also include the num_channels
-        shape = tuple(shape) + tuple([self.header.num_channels])
-
-        data = np.zeros(shape=shape, dtype=self.header.dtype_per_channel)
+        shape_with_num_channels = shape + (self.header.num_channels,)
+        data = np.zeros(
+            shape=shape_with_num_channels, dtype=self.header.dtype_per_channel
+        )
         for (
             xyz,
             _,
             offset_in_output_data,
             offset_in_input_data,
-        ) in self.calculate_relevant_slices(off, shape):
-            x, y, z = xyz
+        ) in self.calculate_relevant_slices(off, shape_with_num_channels):
+            _, _, z = xyz
             z_index_in_data = z - off[2]
 
             if xyz in self.tiffs:
@@ -201,8 +227,10 @@ class TiffMag:
                 loaded_data = np.array(
                     self.tiffs[xyz].read(), self.header.dtype_per_channel
                 )[
-                    offset_in_output_data[0] : offset_in_output_data[0] + shape[0],
-                    offset_in_output_data[1] : offset_in_output_data[1] + shape[1],
+                    offset_in_output_data[0] : offset_in_output_data[0]
+                    + shape_with_num_channels[0],
+                    offset_in_output_data[1] : offset_in_output_data[1]
+                    + shape_with_num_channels[1],
                 ]
 
                 index_slice = [
@@ -226,7 +254,7 @@ class TiffMag:
         data = np.moveaxis(data, -1, 0)
         return data
 
-    def write(self, off, data):
+    def write(self, off: Tuple[int, int, int], data: np.ndarray) -> None:
         if not len(data.shape) == 3:
             # reformat array to have the channels as the first index (similar to wkw)
             # this is only necessary if the data has a dedicated dimensions for the num_channels
@@ -265,20 +293,28 @@ class TiffMag:
 
             self.tiffs[xyz].merge_with_image(pixel_data, offset_in_output_data)
 
-    def compress(self, dst_path: str, compress_files: bool = False):
-        raise NotImplementedError
-
-    def list_files(self):
+    def list_files(self) -> Iterator[str]:
         _, file_extension = os.path.splitext(self.header.pattern)
         file_paths = list(iglob(os.path.join(self.root, "*" + file_extension)))
 
         for file_path in file_paths:
             yield os.path.relpath(os.path.normpath(file_path), self.root)
 
-    def close(self):
+    def close(self) -> None:
         return
 
-    def calculate_relevant_slices(self, offset, shape):
+    def calculate_relevant_slices(
+        self,
+        offset: Tuple[int, int, int],
+        shape: Union[Tuple[int, int, int, int], Tuple[int, int, int]],
+    ) -> Iterator[
+        Tuple[
+            Tuple[Optional[int], Optional[int], int],
+            Tuple[int, ...],
+            Tuple[int, int],
+            Tuple[int, int],
+        ]
+    ]:
         """
         The purpose of this method is to find out which tiles need to be touched.
         Each tile is specified by its (x, y, z)-dimensions.
@@ -297,17 +333,17 @@ class TiffMag:
 
         if tile_size is None:
             x_first_index = None
-            x_indices = [None]
+            x_indices: List[Union[int, None]] = [None]
             y_first_index = None
-            y_indices = [None]
+            y_indices: List[Union[int, None]] = [None]
         else:
             x_first_index = offset[0] // tile_size[0]  # floor division
             x_last_index = np.math.ceil(max_indices[0] / tile_size[0])
-            x_indices = range(x_first_index, x_last_index)
+            x_indices = list(range(x_first_index, x_last_index))
 
             y_first_index = offset[1] // tile_size[1]  # floor division
             y_last_index = np.math.ceil(max_indices[1] / tile_size[1])
-            y_indices = range(y_first_index, y_last_index)
+            y_indices = list(range(y_first_index, y_last_index))
 
         for x in x_indices:
             for y in y_indices:
@@ -331,26 +367,25 @@ class TiffMag:
 
                         offset_in_input_data = shape_top_left_corner - offset[0:2]
                         offset_in_output_data = tuple(
-                            (offset[0:2] - shape_top_left_corner)
+                            (np.array(offset[0:2]) - shape_top_left_corner)
                             * np.equal((x, y), (x_first_index, y_first_index))
                         )
-                        tile_shape = tuple(
-                            shape_bottom_right - shape_top_left_corner
-                        ) + tuple(shape[3:4])
-
-                    yield tuple(
-                        (
-                            (x, y, z),
-                            tile_shape,
-                            offset_in_output_data,
-                            offset_in_input_data,
+                        tile_shape = (
+                            tuple(shape_bottom_right - shape_top_left_corner)
+                            + shape[3:4]
                         )
+
+                    yield (
+                        (x, y, z),
+                        tile_shape,
+                        cast(Tuple[int, int], offset_in_output_data),
+                        offset_in_input_data,
                     )
 
     def has_only_one_channel(self) -> bool:
         return self.header.num_channels == 1
 
-    def assert_correct_data_format(self, data):
+    def assert_correct_data_format(self, data: np.ndarray) -> None:
         if self.has_only_one_channel():
             if not len(data.shape) == 3:
                 raise AttributeError(
@@ -370,38 +405,31 @@ class TiffMag:
                 f"The type of the provided data does not match the expected type. (Expected np.array of type {self.header.dtype_per_channel.name})"
             )
 
-    def get_file_name_for_layer(self, xyz) -> str:
+    def get_file_name_for_layer(
+        self, xyz: Tuple[Optional[int], Optional[int], Optional[int]]
+    ) -> str:
         x, y, z = xyz
         return os.path.join(self.root, to_file_name(self.header.pattern, x, y, z))
 
     @staticmethod
-    def open(root: str, header=None):
+    def open(root: str, header: TiffMagHeader = None) -> "TiffMag":
         if header is None:
             header = TiffMagHeader()
         return TiffMag(root, header)
 
-    def __enter__(self):
+    def __enter__(self) -> "TiffMag":
         return self
 
-    def __exit__(self, type, value, tb):
+    def __exit__(
+        self,
+        _type: Optional[Type[BaseException]],
+        _value: Optional[BaseException],
+        _tb: Optional[TracebackType],
+    ) -> None:
         self.close()
 
 
-class TiffMagHeader:
-    def __init__(
-        self,
-        pattern="{zzzzz}.tif",
-        dtype_per_channel=np.dtype("uint8"),
-        num_channels=1,
-        tile_size=(32, 32),
-    ):
-        self.pattern = pattern
-        self.dtype_per_channel = np.dtype(dtype_per_channel)
-        self.num_channels = num_channels
-        self.tile_size = tile_size
-
-
-def transpose_for_skimage(data):
+def transpose_for_skimage(data: np.ndarray) -> np.ndarray:
     if len(data.shape) == 2:
         return data.transpose()
     elif len(data.shape) == 3:
@@ -411,28 +439,30 @@ def transpose_for_skimage(data):
 
 
 class TiffReader:
-    def __init__(self, file_name):
+    def __init__(self, file_name: str):
         self.file_name = file_name
 
     @classmethod
-    def init_tiff(cls, pixels, file_name):
+    def init_tiff(cls, pixels: np.ndarray, file_name: str) -> "TiffReader":
         tr = TiffReader(file_name)
         tr.write(pixels)
         return tr
 
     @classmethod
-    def open(cls, file_name):
+    def open(cls, file_name: str) -> "TiffReader":
         return cls(file_name)
 
     def read(self) -> np.array:
         data = io.imread(self.file_name)
         return transpose_for_skimage(data)
 
-    def write(self, pixels):
+    def write(self, pixels: np.ndarray) -> None:
         os.makedirs(os.path.dirname(self.file_name), exist_ok=True)
         io.imsave(self.file_name, transpose_for_skimage(pixels), check_contrast=False)
 
-    def merge_with_image(self, foreground_pixels, offset):
+    def merge_with_image(
+        self, foreground_pixels: np.ndarray, offset: Tuple[int, int]
+    ) -> None:
         background_pixels = self.read()
         bg_shape = background_pixels.shape
         fg_shape = foreground_pixels.shape
