@@ -2,11 +2,14 @@ import logging
 import time
 from argparse import ArgumentParser
 from pathlib import Path
+from sklearn.preprocessing import LabelEncoder
+from typing import Tuple, Optional, Union, cast
 
 import nibabel as nib
 import numpy as np
 
 from wkcuber.api.Dataset import TiffDataset, WKDataset
+from wkcuber.api.bounding_box import BoundingBox
 from wkcuber.utils import (
     DEFAULT_WKW_FILE_LEN,
     DEFAULT_WKW_VOXELS_PER_BLOCK,
@@ -18,7 +21,7 @@ from wkcuber.utils import (
 )
 
 
-def create_parser():
+def create_parser() -> ArgumentParser:
     parser = ArgumentParser()
 
     parser.add_argument(
@@ -38,6 +41,15 @@ def create_parser():
         default="color",
     )
 
+    parser.add_argument(
+        "--is_segmentation_layer",
+        "-s",
+        help="When converting one layer, signals whether layer is segmentation layer. "
+        "When converting a folder, this option is ignored",
+        default=False,
+        action="store_true",
+    )
+
     parser.add_argument(
         "--dtype", "-d", help="Target datatype (e.g. uint8, uint16).", default="uint8"
     )
@@ -88,13 +100,17 @@ def create_parser():
 
 
 def to_target_datatype(
-    data: np.ndarray, target_dtype, is_probably_binary: bool
+    data: np.ndarray, target_dtype: str, is_segmentation_layer: bool
 ) -> np.ndarray:
-    if is_probably_binary:
-        logging.info(
-            f"Casting directly to {target_dtype}, as input seems to be binary."
+
+    if is_segmentation_layer:
+        original_shape = data.shape
+        label_encoder = LabelEncoder()
+        return (
+            label_encoder.fit_transform(data.ravel())
+            .reshape(original_shape)
+            .astype(np.dtype(target_dtype))
         )
-        return data.astype(np.dtype(target_dtype))
 
     if data.dtype == np.dtype("float32"):
         factor = data.max()
@@ -111,18 +127,18 @@ def to_target_datatype(
 
 
 def convert_nifti(
-    source_nifti_path,
-    target_path,
-    layer_name,
-    dtype,
-    scale,
-    mag=1,
-    file_len=DEFAULT_WKW_FILE_LEN,
-    bbox_to_enforce=None,
-    write_tiff=False,
-    use_orientation_header=False,
-    flip_axes=None,
-):
+    source_nifti_path: Path,
+    target_path: Path,
+    layer_name: str,
+    dtype: str,
+    scale: Tuple[float, ...],
+    is_segmentation_layer: bool = False,
+    file_len: int = DEFAULT_WKW_FILE_LEN,
+    bbox_to_enforce: BoundingBox = None,
+    write_tiff: bool = False,
+    use_orientation_header: bool = False,
+    flip_axes: Optional[Union[int, Tuple[int, ...]]] = None,
+) -> None:
     voxels_per_cube = file_len * DEFAULT_WKW_VOXELS_PER_BLOCK
     ref_time = time.time()
 
@@ -136,16 +152,12 @@ def convert_nifti(
 
     cube_data = np.array(source_nifti.get_fdata())
 
-    is_probably_binary = np.unique(cube_data).shape[0] == 2
-    assume_segmentation_layer = (
-        False
-    )  # Since webKnossos does not support multiple segmention layers, this is hardcoded to False right now.
     max_cell_id_args = (
         {"largest_segment_id": int(np.max(cube_data) + 1)}
-        if assume_segmentation_layer
+        if is_segmentation_layer
         else {}
     )
-    category_type = "segmentation" if assume_segmentation_layer else "color"
+    category_type = "segmentation" if is_segmentation_layer else "color"
     logging.debug(f"Assuming {category_type} as layer type for {layer_name}")
 
     if len(source_nifti.shape) == 3:
@@ -172,8 +184,9 @@ def convert_nifti(
 
     if scale is None:
         scale = tuple(map(float, source_nifti.header["pixdim"][:3]))
+
     logging.info(f"Using scale: {scale}")
-    cube_data = to_target_datatype(cube_data, dtype, is_probably_binary)
+    cube_data = to_target_datatype(cube_data, dtype, is_segmentation_layer)
 
     # everything needs to be padded to
     if bbox_to_enforce is not None:
@@ -199,8 +212,10 @@ def convert_nifti(
     )
 
     if write_tiff:
-        ds = TiffDataset.get_or_create(target_path, scale=scale or (1, 1, 1))
-        layer = ds.get_or_add_layer(
+        tiff_ds = TiffDataset.get_or_create(
+            target_path, scale=cast(Tuple[float, float, float], scale or (1, 1, 1))
+        )
+        layer = tiff_ds.get_or_add_layer(
             layer_name,
             category_type,
             dtype_per_layer=np.dtype(dtype),
@@ -210,8 +225,10 @@ def convert_nifti(
 
         mag.write(cube_data.squeeze())
     else:
-        ds = WKDataset.get_or_create(target_path, scale=scale or (1, 1, 1))
-        layer = ds.get_or_add_layer(
+        wk_ds = WKDataset.get_or_create(
+            target_path, scale=cast(Tuple[float, float, float], scale or (1, 1, 1))
+        )
+        layer = wk_ds.get_or_add_layer(
             layer_name,
             category_type,
             dtype_per_layer=np.dtype(dtype),
@@ -228,16 +245,16 @@ def convert_nifti(
 
 
 def convert_folder_nifti(
-    source_folder_path,
-    target_path,
-    color_subpath,
-    segmentation_subpath,
-    scale,
-    use_orientation_header=False,
-    bbox_to_enforce=None,
-    write_tiff=False,
-    flip_axes=None,
-):
+    source_folder_path: Path,
+    target_path: Path,
+    color_subpath: str,
+    segmentation_subpath: str,
+    scale: Tuple[float, ...],
+    use_orientation_header: bool = False,
+    bbox_to_enforce: BoundingBox = None,
+    write_tiff: bool = False,
+    flip_axes: Optional[Union[int, Tuple[int, ...]]] = None,
+) -> None:
     paths = list(source_folder_path.rglob("**/*.nii"))
 
     color_path = None
@@ -262,23 +279,49 @@ def convert_folder_nifti(
 
     logging.info("Segmentation file will also use uint8 as a datatype.")
 
-    conversion_args = {
-        "scale": scale,
-        "write_tiff": write_tiff,
-        "bbox_to_enforce": bbox_to_enforce,
-        "use_orientation_header": use_orientation_header,
-        "flip_axes": flip_axes,
-    }
     for path in paths:
         if path == color_path:
-            convert_nifti(path, target_path, "color", "uint8", **conversion_args)
+            convert_nifti(
+                path,
+                target_path,
+                "color",
+                "uint8",
+                scale,
+                is_segmentation_layer=False,
+                write_tiff=write_tiff,
+                bbox_to_enforce=bbox_to_enforce,
+                use_orientation_header=use_orientation_header,
+                flip_axes=flip_axes,
+            )
         elif path == segmentation_path:
-            convert_nifti(path, target_path, "segmentation", "uint8", **conversion_args)
+            convert_nifti(
+                path,
+                target_path,
+                "segmentation",
+                "uint8",
+                scale,
+                is_segmentation_layer=True,
+                write_tiff=write_tiff,
+                bbox_to_enforce=bbox_to_enforce,
+                use_orientation_header=use_orientation_header,
+                flip_axes=flip_axes,
+            )
         else:
-            convert_nifti(path, target_path, path.stem, "uint8", **conversion_args)
+            convert_nifti(
+                path,
+                target_path,
+                path.stem,
+                "uint8",
+                scale,
+                is_segmentation_layer=False,
+                write_tiff=write_tiff,
+                bbox_to_enforce=bbox_to_enforce,
+                use_orientation_header=use_orientation_header,
+                flip_axes=flip_axes,
+            )
 
 
-def main():
+def main() -> None:
     args = create_parser().parse_args()
     setup_logging(args)
 
@@ -314,6 +357,7 @@ def main():
             Path(args.target_path),
             args.layer_name,
             args.dtype,
+            is_segmentation_layer=args.is_segmentation_layer,
             **conversion_args,
         )
 
